@@ -1,8 +1,12 @@
-"""Core data classes."""
+"""Core data classes and functions."""
 
+import copy
+import functools
 import numpy as np
 import os
-import functools
+import time
+from logging import getLogger, Formatter, StreamHandler
+from typing import Any, List, Optional, Tuple
 
 
 class Defaults:
@@ -11,8 +15,7 @@ class Defaults:
     that may be used in object initialisation.
 
     Implementation of the singleton design pattern is based on:
-    https://python-3-patterns-idioms-test.readthedocs.io
-           /en/latest/Singleton.html
+    https://python-3-patterns-idioms-test.readthedocs.io/en/latest/Singleton.html
     """
 
     # Define the single instance as a class attribute
@@ -21,29 +24,31 @@ class Defaults:
     # Create single instance in inner class
     class __Defaults:
 
-        # Define instance attributes based on opts dictionary
-        def __init__(self, opts={}):
-            for key, value in opts.items():
-                setattr(self, key, value)
+        def __init__(self, opts: Optional[dict] = None):
+            """Define instance attributes based on opts dictionary."""
+
+            if opts:
+                for key, value in opts.items():
+                    setattr(self, key, value)
 
-        # Allow for printing instance attributes
         def __repr__(self):
-            out_list = []
+            """Print instance attributes."""
+
+            out = []
             for key, value in sorted(self.__dict__.items()):
                 out_list.append(f"{key}: {value}")
-            out_string = "\n".join(out_list)
-            return out_string
+            return "\n".join(out)
 
-    def __init__(self, opts={}, reset=False):
+    def __init__(self, opts: Optional[dict] = None, reset: bool = False):
         """
         Constructor of Defaults singleton class.
 
         Parameters
         ----------
-        opts : dict, default={}
+        opts: dict, default={}
             Dictionary of attribute-value pairs.
 
-        reset : bool, default=False
+        reset: bool, default=False
             If True, delete all pre-existing instance attributes before
             adding attributes and values from opts dictionary.
             If False, don't delete pre-existing instance attributes,
@@ -55,24 +60,38 @@ class Defaults:
         else:
             if reset:
                 Defaults.instance.__dict__ = {}
-            for key, value in opts.items():
-                setattr(Defaults.instance, key, value)
+            if opts:
+                for key, value in opts.items():
+                    setattr(Defaults.instance, key, value)
+
+    def __getattr__(self, name: str):
+        """Get instance attributes."""
 
-    # Allow for getting instance attributes
-    def __getattr__(self, name):
         return getattr(self.instance, name)
 
-    # Allow for setting instance attributes
-    def __setattr__(self, name, value):
+    def __setattr__(self, name: str, value: Any):
+        """Set instance attributes."""
+
         return setattr(self.instance, name, value)
 
-    # Allow for printing instance attributes
     def __repr__(self):
+        """Print instance attributes."""
+
         return self.instance.__repr__()
 
 
+# Initialise default parameter values:
+Defaults()
+
+# Depth to which recursion is performed when printing instances
+# of classes that inherit from the Data class.
 Defaults({"print_depth": 0})
 
+# Severity level for event logging.
+# Defined values are: 'NOTSET' (0), 'DEBUG' (10), 'INFO' (20),
+# 'WARNING' (30), 'ERROR' (40), 'CRITICAL' (50)
+Defaults({"log_level": "WARNING"})
+
 
 class Data:
     """
@@ -85,14 +104,14 @@ class Data:
     attributes and values.
     """
 
-    def __init__(self, opts={}, **kwargs):
+    def __init__(self, opts: Optional[dict] = None, **kwargs):
         """
         Constructor of Data class, allowing initialisation of an
         arbitrary set of attributes.
 
         Parameters
         ----------
-        opts : dict, default={}
+        opts: dict, default={}
             Dictionary to be used in setting instance attributes
             (dictionary keys) and their initial values.
 
@@ -101,20 +120,20 @@ class Data:
             and their initial values.
         """
 
-        for key, value in opts.items():
-            setattr(self, key, value)
+        if opts:
+            for key, value in opts.items():
+                setattr(self, key, value)
         for key, value in kwargs.items():
             setattr(self, key, value)
-        return None
 
-    def __repr__(self, depth=None):
+    def __repr__(self, depth: Optional[int] = None) -> str:
         """
         Create string recursively listing attributes and values.
 
         Parameters
         ----------
 
-        depth : integer/None, default=None
+        depth: integer/None, default=None
             Depth to which recursion is performed.
             If the value is None, depth is set to the value
             of the object's print_depth property, if defined,
@@ -124,7 +143,7 @@ class Data:
         if depth is None:
             depth = self.get_print_depth()
 
-        out_list = [f"\n{self.__class__.__name__}", "{"]
+        out = [f"\n{self.__class__.__name__}", "{"]
 
         # Loop over attributes, with different treatment
         # depending on whether attribute value is a list.
@@ -134,15 +153,29 @@ class Data:
         # the instance's __repr__() method with depth decreased
         # by 1, or (depth less than 1) is the class representation.
         for key in sorted(self.__dict__):
+
+            # Ignore private attributes 
+            if key.startswith("_"):
+                continue
+
             item = self.__dict__[key]
-            if isinstance(item, list):
+
+            # Handle printing of numpy arrays
+            if isinstance(item, np.ndarray):
+                value_string = f"{item.shape} array"
+
+            # Handle printing of lists
+            elif isinstance(item, list):
                 items = item
                 n = len(items)
                 if n:
                     if depth > 0:
                         value_string = "["
                         for i, item in enumerate(items):
-                            item_string = item.__repr__(depth=(depth - 1))
+                            try:
+                                item_string = item.__repr__(depth=(depth - 1))
+                            except TypeError:
+                                item_string = item.__repr__()
                             comma = "," if (i + 1 < n) else ""
                             value_string = f"{value_string} {item_string}{comma}"
                         value_string = f"{value_string}]"
@@ -150,6 +183,29 @@ class Data:
                         value_string = f"[{n} * {item[0].__class__}]"
                 else:
                     value_string = "[]"
+
+            # Handle printing of dicts
+            elif isinstance(item, dict):
+                items = item
+                n = len(items)
+                if n:
+                    if depth > 0:
+                        value_string = "{"
+                        for i, (key, value) in enumerate(items.items()):
+                            item_string = "{key}: "
+                            try:
+                                item_string += item.__repr__(depth=(depth - 1))
+                            except TypeError:
+                                item_string += item.__repr__()
+                            comma = "," if (i + 1 < n) else ""
+                            value_string = f"{value_string} {item_string}{comma}"
+                        value_string = f"{{{value_string}}}"
+                    else:
+                        value_string = f"{{{n} * keys of type {list(item.keys())[0].__class__}}}"
+                else:
+                    value_string = "{}"
+
+            # Handle printing of nested Data objects
             else:
                 if issubclass(item.__class__, Data):
                     if depth > 0:
@@ -158,12 +214,46 @@ class Data:
                         value_string = f"{item.__class__}"
                 else:
                     value_string = item.__repr__()
-            out_list.append(f"  {key} : {value_string} ")
-        out_list.append("}")
-        out_string = "\n".join(out_list)
-        return out_string
+            out.append(f"  {key}: {value_string} ")
 
-    def get_dict(self):
+        out.append("}")
+        return "\n".join(out)
+
+    def clone(self):
+        """
+        Return a clone of the Data object containing copies of any 
+        lists, dicts, or arrays owned by the object.
+        """
+
+        clone = copy.copy(self)
+        self.clone_attrs(clone)
+        return clone
+
+    def clone_attrs(self, obj):
+        """
+        Apply class attributes from self to a new object, making copies 
+        of any lists, dicts, or numpy arrays.
+        """
+
+        for attr in dir(self):
+
+            # Don't copy private variables
+            if attr.startswith("__"):
+                continue
+
+            # Don't copy methods
+            if callable(getattr(self, attr)):
+                continue
+            
+            # Make new copy of lists/dicts/arrays
+            if type(getattr(self, attr)) in [list, dict, np.ndarray]:
+                setattr(obj, attr, copy.copy(getattr(self, attr)))
+
+            # Otherwise, copy reference to attribute
+            else:
+                setattr(obj, attr, getattr(self, attr))
+
+    def get_dict(self) -> dict:
         """
         Return a nested dictionary of object attributes (dictionary keys)
         and their values.
@@ -178,7 +268,7 @@ class Data:
 
         return objects
 
-    def get_print_depth(self):
+    def get_print_depth(self) -> int:
         """
         Retrieve the value of the object's print depth,
         setting an initial value if not previously defined.
@@ -188,7 +278,7 @@ class Data:
             self.set_print_depth()
         return self.print_depth
 
-    def print(self, depth=None):
+    def print(self, depth: Optional[int] = None):
         """
         Convenience method for recursively printing
         object attributes and values, with recursion
@@ -197,23 +287,22 @@ class Data:
         Parameters
         ----------
 
-        depth : integer/None, default=None
+        depth: integer/None, default=None
             Depth to which recursion is performed.
             If the value is None, depth is set in the
             __repr__() method.
         """
 
         print(self.__repr__(depth))
-        return None
 
-    def set_print_depth(self, depth=None):
+    def set_print_depth(self, depth: Optional[int] = None):
         """
         Set the object's print depth.
 
         Parameters
         ----------
 
-        depth : integer/None, default=None
+        depth: integer/None, default=None
             Depth to which recursion is performed.
             If the value is None, the object's print depth is
             set to the value of Defaults().print_depth.
@@ -222,18 +311,20 @@ class Data:
         if depth is None:
             depth = Defaults().print_depth
         self.print_depth = depth
-        return None
 
 
 class PathData(Data):
-    """Data with and associated directory; has the ability to
+    """Data with an associated path or directory; has the ability to
     extract a list of dated objects from within this directory."""
 
     def __init__(self, path=""):
         self.path = fullpath(path)
         self.subdir = ""
 
-    def get_dated_objects(self, dtype, subdir="", **kwargs):
+    def get_dated_objects(self, 
+                          dtype: type, 
+                          subdir: str = "", 
+                          **kwargs) -> List[Any]:
         """Create list of objects of a given type, <dtype>, inside own
         directory, or inside own directory + <subdir> if given."""
 
@@ -241,8 +332,7 @@ class PathData(Data):
         objs = []
         path = os.path.join(self.path, subdir)
         if os.path.isdir(path):
-            filenames = os.listdir(path)
-            for filename in filenames:
+            for filename in os.listdir(path):
                 if is_timestamp(filename):
                     filepath = os.path.join(path, filename)
                     try:
@@ -264,7 +354,12 @@ class Dated(PathData):
     """PathData with an associated date and time, which can be used for
     sorting multiple Dateds."""
 
-    def __init__(self, path=""):
+    def __init__(self, path: str = "", auto_timestamp=False):
+        """
+        Initialise dated object from a path and assign its timestamp. If 
+        no valid timestamp is found within the path string, it will be set 
+        automatically from the current date and time if auto_timestamp is True.
+        """
 
         PathData.__init__(self, path)
 
@@ -281,9 +376,17 @@ class Dated(PathData):
             except ValueError:
                 self.date, self.time = (None, None)
 
+        # Set date and time from current time
+        if not self.date and auto_timestamp:
+            self.date = time.strftime("%Y%m%d")
+            self.time = time.strftime("%H%M%S")
+
+        # Make full timestamp string
         self.timestamp = f"{self.date}_{self.time}"
 
-    def in_date_interval(self, min_date=None, max_date=None):
+    def in_date_interval(self, 
+                         min_date: Optional[str] = None, 
+                         max_date: Optional[str] = None) -> bool:
         """Check whether own date falls within an interval."""
 
         if min_date:
@@ -311,7 +414,6 @@ class Dated(PathData):
         return self.date > other.date
 
     def __le__(self, other):
-        return self
         if self.date == other.date:
             return self.time < other.time
         return self.date < other.date
@@ -320,7 +422,7 @@ class Dated(PathData):
 class MachineData(Dated):
     """Dated object with an associated machine name."""
 
-    def __init__(self, path=""):
+    def __init__(self, path: str = ""):
         Dated.__init__(self, path)
         self.machine = os.path.basename(os.path.dirname(path))
 
@@ -328,19 +430,24 @@ class MachineData(Dated):
 class Archive(Dated):
     """Dated object associated with multiple files."""
 
-    def __init__(self, path="", allow_dirs=False):
+    def __init__(self, path: str = "", allow_dirs: bool = False):
 
         Dated.__init__(self, path)
+
+        # Find names of files within the directory
         self.files = []
         try:
             filenames = os.listdir(self.path)
         except OSError:
             filenames = []
+
         for filename in filenames:
 
             # Disregard hidden files
             if not filename.startswith("."):
                 filepath = os.path.join(self.path, filename)
+
+                # Disregard directories unless allow_dirs is True
                 if not os.path.isdir(filepath) or allow_dirs:
                     self.files.append(File(path=filepath))
 
@@ -351,23 +458,9 @@ class File(Dated):
     """File with an associated date. Files can be sorted based on their
     filenames."""
 
-    def __init__(self, path=""):
+    def __init__(self, path: str = ""):
         Dated.__init__(self, path)
 
-    def __cmp__(self, other):
-
-        result = Dated.__cmp__(self, other)
-        if not result:
-            self_basename = os.path.basename(self.path)
-            other_basename = os.path.basename(other.path)
-            basenames = [self_basename, other_basename]
-            basenames.sort(key=alphanumeric)
-            if basenames[0] == self_basename:
-                result = -1
-            else:
-                result = 1
-        return result
-
     def __eq__(self, other):
         return self.path == other.path
 
@@ -395,7 +488,7 @@ class File(Dated):
         return result
 
 
-def alphanumeric(in_str=""):
+def alphanumeric(in_str: str = "") -> List[str]:
     """Function that can be passed as value for list sort() method
     to have alphanumeric (natural) sorting"""
 
@@ -411,7 +504,7 @@ def alphanumeric(in_str=""):
     return elements
 
 
-def fullpath(path=""):
+def fullpath(path: str = "") -> str:
     """Evaluate full path, expanding '~', environment variables, and
     symbolic links."""
 
@@ -422,8 +515,28 @@ def fullpath(path=""):
         expanded = os.path.realpath(tmp)
     return expanded
 
+def get_logger(name="", log_level=None):
+    """
+    Retrieve named event logger.
 
-def get_time_and_date(timestamp=""):
+    Parameters
+    ----------
+    name: string, default=""
+        Name of logger (see documentation of logging module)
+    log_level: string/integer/None, default=None
+        Severity level for event logging.  If the value is None,
+        log_level is set to the value of Defaults().log_level.
+    """
+    formatter = Formatter("%(name)s - %(levelname)s - %(message)s")
+    handler = StreamHandler()
+    handler.setFormatter(formatter)
+    logger = getLogger(name)
+    if not logger.handlers:
+        logger.addHandler(handler)
+    logger.setLevel(log_level)
+    return logger
+
+def get_time_and_date(timestamp: str = "") -> Tuple[str, str]:
     """Extract time and date separately from timestamp."""
 
     time_date = (None, None)
@@ -438,14 +551,14 @@ def get_time_and_date(timestamp=""):
         i1 = timestamp.find("_")
         i2 = timestamp.rfind(".")
         if (-1 != i1) and (-1 != i2):
-            bitstamp = timestamp[i1 + 1 : i2]
+            bitstamp = timestamp[i1 + 1: i2]
             if is_timestamp(bitstamp):
                 time_date = tuple(bitstamp.split("_"))
 
     return time_date
 
 
-def is_timestamp(string=""):
+def is_timestamp(string: str = "") -> bool:
     """Check whether a string is a valid timestamp."""
 
     valid = True
@@ -468,7 +581,7 @@ def is_timestamp(string=""):
     return valid
 
 
-def is_list(var):
+def is_list(var: Any) -> bool:
     """Check whether a variable is a list, tuple, or array."""
 
     is_a_list = False
@@ -478,7 +591,7 @@ def is_list(var):
     return is_a_list
 
 
-def to_three(val):
+def to_three(val: Any) -> Optional[List]:
     """Ensure that a value is a list of three items."""
 
     if val is None:
@@ -486,6 +599,12 @@ def to_three(val):
     if is_list(val):
         if not len(val) == 3:
             print(f"Warning: {val} should be a list containing 3 items!")
-        return val
+        return list(val)
     elif not is_list(val):
         return [val, val, val]
+
+
+def generate_timestamp() -> str:
+    '''Make timestamp from the current time.'''
+
+    return time.strftime('%Y%m%d_%H%M%S')
