@@ -1334,7 +1334,7 @@ class ROI(skrt.image.Image):
         kwargs["ax"] = self.ax
         kwargs["color"] = roi2_color
         kwargs["include_image"] = False
-        other.plot(show=show, **kwargs)
+        other.plot(show=False, **kwargs)
         self.ax.set_title(self.get_comparison_name(other))
 
         if legend:
@@ -1349,9 +1349,13 @@ class ROI(skrt.image.Image):
                 mpatches.Patch(color=roi2_color, label=roi2_name),
             ]
             self.ax.legend(
-                handles=handles, framealpha=1, facecolor="white", loc="lower left"
+                handles=handles, framealpha=1, facecolor="white", 
+                loc="lower left"
             )
 
+        if show:
+            plt.show()
+
         if save_as:
             plt.tight_layout()
             self.fig.savefig(save_as)
@@ -1531,7 +1535,7 @@ class StructureSet(skrt.core.Archive):
         # Expand any directories
         sources_expanded = []
         for source in sources:
-            if os.path.isdir(source):
+            if isinstance(source, str) and os.path.isdir(source):
                 sources_expanded.extend(
                     [os.path.join(source, file) for file in os.listdir(source)]
                 )
@@ -1574,6 +1578,9 @@ class StructureSet(skrt.core.Archive):
 
         self.rename_rois()
         self.filter_rois()
+        for roi in self.rois:
+            roi.structure_set = self
+
         self.loaded = True
 
     def reset(self):
@@ -1675,7 +1682,7 @@ class StructureSet(skrt.core.Archive):
         if not skrt.core.is_list(sources):
             sources = [sources]
         self.sources.extend(sources)
-        self.load_rois(sources)
+        self.load(sources)
 
     def add_roi(self, source, **kwargs):
         """Add a single ROI with  optional kwargs."""
@@ -1685,9 +1692,10 @@ class StructureSet(skrt.core.Archive):
             roi = source
         else:
             roi = ROI(source, **kwargs)
+        roi.structure_set = self
         self.rois.append(roi)
 
-    def copy(
+    def filter(
         self,
         names=None,
         name=None,
@@ -1696,7 +1704,7 @@ class StructureSet(skrt.core.Archive):
         keep_renamed_only=False,
     ):
         """Create a copy of this structure set with ROIs optionally
-        renamed or filtered."""
+        renamed or filtered. Returns a new StructureSet object."""
 
         if not hasattr(self, "n_copies"):
             self.n_copies = 1
@@ -1722,11 +1730,29 @@ class StructureSet(skrt.core.Archive):
 
         return ss
 
-    def get_rois(self):
-        """Get list of ROI objects."""
+    def get_rois(self, names=None):
+        """Get list of ROI objects If <names> is given, only the ROIs with
+        those names will be returned."""
 
         self.load()
-        return self.rois
+        if names is None:
+            return self.rois
+
+        rois = []
+        for name in names:
+            roi = self.get_roi(name)
+            if roi is not None:
+                rois.append(roi)
+        return rois
+
+    def get_rois_wildcard(self, wildcard):
+        """Return list of ROIs matching a wildcard expression."""
+
+        rois = []
+        for roi in self.get_rois():
+            if fnmatch.fnmatch(roi.name, wildcard):
+                rois.append(roi)
+        return rois
 
     def get_roi_names(self, original=False):
         """
@@ -1734,17 +1760,15 @@ class StructureSet(skrt.core.Archive):
         get the original names of the ROIs.
         """
 
-        self.load()
         if not original:
-            return [s.name for s in self.rois]
+            return [s.name for s in self.get_rois()]
         else:
-            return [s.original_name for s in self.rois]
+            return [s.original_name for s in self.get_rois()]
 
     def get_roi_dict(self):
         """Get dict of ROI names and objects."""
 
-        self.load()
-        return {s.name: s for s in self.rois}
+        return {s.name: s for s in self.get_rois()}
 
     def get_roi(self, name):
         """Get an ROI with a specific name."""
@@ -1937,12 +1961,14 @@ def load_rois_dicom(path, names=None):
 
     # Load dicom object
     try:
-        ds = pydicom.read_file(path)
+        ds = pydicom.read_file(path, force=True)
     except pydicom.errors.InvalidDicomError:
         return []
+    if not hasattr(ds, "SOPClassUID"):
+        return []
     if not (ds.SOPClassUID == "1.2.840.10008.5.1.4.1.1.481.3"):
         print(f"Warning: {path} is not a DICOM structure set file!")
-        return
+        return []
 
     # Get ROI names
     seq = get_dicom_sequence(ds, "StructureSetROI")
